#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import os
import pathlib
from ..logging import log

def split_pdf_to_pages(pdf_path, temp_dir):

    log.info(f"Splitting PDF {pdf_path} into pages...")

    pdf_path = pathlib.Path(pdf_path)
    from pypdf import PdfReader, PdfWriter
    pdf = PdfReader(pdf_path)

    log.info(f"PDF file {pdf_path} contains {len(pdf.pages)} pages")

    # Get base name without extension
    basename = os.path.splitext(os.path.basename(pdf_path))[0]

    page_files = []
    
    if len(pdf.pages) == 1:
        #log.debug(f"Only one page in PDF {pdf_path} - sending back")
        return [str(pdf_path)]
    
    for page in range(len(pdf.pages)):
        pdf_writer = PdfWriter()
        pdf_writer.add_page(pdf.pages[page])
        
        page_str = "{:02d}".format(page + 1)  
        output_filename = pathlib.Path(temp_dir, f'{basename}_p{page_str}.pdf')

        with open(output_filename, 'wb') as out:
            pdf_writer.write(out)

        log.info(f'Created PDF page: {output_filename}')
        page_files.append(str(output_filename))

    log.info(f"Split PDF {pdf_path} into {len(page_files)} pages...")
    return page_files

def read_pdf_file(pdf_path, metadata):
    from langchain.schema import Document
    from pypdf import PdfReader
    log.info(f"Trying local PDF parsing.  Reading PDF {pdf_path}...")

    pdf_path = pathlib.Path(pdf_path)
    
    pdf = PdfReader(pdf_path)

    try:
        text = ""
        for page in pdf.pages:
            text += page.extract_text() + "\n"
    except Exception as err:
        log.warning(f"Could not extract PDF via pypdf ERROR - {str(err)}")
        return None
    
    if len(text) < 10:
        log.info(f"Could not read PDF {pdf_path} via pypdf - too short, only got {text}")
        return None
    
    log.info(f"Successfully read PDF {pdf_path}...")
    return Document(page_content=text, metadata=metadata)


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Any, Dict, List, Union
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from langchain.schema import LLMResult

import threading
import re
from ..logging import log



class ContentBuffer:
    def __init__(self):
        self.content = ""
        log.debug("Content buffer initialized")
    
    def write(self, text: str):
        self.content += text
        log.debug(f"Written {text} to buffer")
    
    def read(self) -> str:
        log.debug(f"Read content from buffer")    
        return self.content

    def clear(self):
        log.debug(f"Clearing content buffer")
        self.content = ""
    

class BufferStreamingStdOutCallbackHandler(StreamingStdOutCallbackHandler):
    def __init__(self, content_buffer: ContentBuffer, tokens: str = ".?!\n", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.content_buffer = content_buffer

        self.tokens = tokens
        self.buffer = ""
        self.stream_finished = threading.Event()
        self.in_code_block = False
        self.in_question_block = False
        self.question_buffer = ""
        log.info("Starting to stream LLM")

    def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        log.debug(f"on_llm_new_token: {token}")

        self.buffer += token

        # Toggle the code block flag if the delimiter is encountered
        if '```' in token:
            self.in_code_block = not self.in_code_block

        # Process the buffer if not inside a code block
        if not self.in_code_block:
            self._process_buffer()

    def _process_buffer(self):

        # Check for the last occurrence of a newline followed by a numbered list pattern
        matches = list(re.finditer(r'\n(\d+\.\s)', self.buffer))
        if matches:
            # If found, write up to the start of the last match, and leave the rest in the buffer
            last_match = matches[-1]
            start_of_last_match = last_match.start() + 1  # Include the newline in the split
            self.content_buffer.write(self.buffer[:start_of_last_match])
            self.buffer = self.buffer[start_of_last_match:]
        else:
            # If not found, and the buffer ends with one of the specified ending tokens, write the entire buffer
            if any(self.buffer.endswith(t) for t in self.tokens):
                self.content_buffer.write(self.buffer)
                self.buffer = ""

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:

        if self.buffer:
            # Process the remaining buffer content
            self.content_buffer.write(self.buffer)
            self.buffer = "" # Clear the remaining buffer
            log.info("Flushing remaining LLM response buffer")

        self.stream_finished.set() # Set the flag to signal that the stream has finished
        log.info("Streaming LLM response ended successfully")


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import json
import yaml
from google.cloud import storage

def fetch_config(bucket_name, blob_name):

    from ..logging import log
    

    storage_client = storage.Client()

    bucket = storage_client.bucket(bucket_name)
    blob = storage.Blob(blob_name, bucket)

    # Check if the file exists
    if not blob.exists():
        log.info(f"The blob {blob_name} does not exist in the bucket {bucket_name}")
        return None

    # Download the file to a local file
    blob.download_to_filename(blob_name)

    # Get the blob's updated time
    updated_time = blob.updated

    return updated_time

def get_module_filepath(filepath):
    from ..logging import log
    
    # Get the root directory of this Python script
    dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    # Build the full filepath by joining the directory with the filename
    filepath = os.path.join(dir_path, filepath)

    log.info(f"Found filepath {filepath}")
    return filepath

# Global cache
config_cache = {}

def load_config(filename: str=None) -> tuple[dict, str]:
    global config_cache
    from ..logging import log
    
    if filename is None:
        filename = os.getenv("_CONFIG_FILE", None)
        if filename is None:
            raise ValueError("No _CONFIG_FILE env value specified")

   # Check the cache first
    if filename in config_cache:
        log.debug(f"Returning cached config for {filename}")
        return config_cache[filename], filename
    
    if os.getenv("_CONFIG_FOLDER"):
        log.debug(f"_CONFIG_FOLDER: {os.getenv('_CONFIG_FOLDER')}")
                  
    # Join the script directory with the filename
    config_path = filename
    config_folder = os.getenv("_CONFIG_FOLDER") if os.getenv("_CONFIG_FOLDER") else os.getcwd()

    config_file = f"{config_folder}/{config_path}"
    log.debug(f"Loading config file {config_file}")

    with open(config_file, 'r') as f:
        if filename.endswith(".json"):
            config = json.load(f)
        elif filename.endswith(".yaml") or filename.endswith(".yml"):
            config = yaml.safe_load(f)
        else:
            raise ValueError(f"Unsupported config file format: {config_file}. The supported formats are JSON and YAML.")

    # Store in cache
    config_cache[filename] = config    
    
    return config, filename

def load_config_key(key: str, vector_name: str, filename: str=None) -> str:
    from ..logging import log
    

    assert isinstance(key, str), f"key must be a string got a {type(key)}"
    assert isinstance(vector_name, str), f"vector_name must be a string, got a {type(vector_name)}"
    
    config, filename = load_config(filename)
    log.info(f"Fetching {key} for {vector_name}")
    llm_config = config.get(vector_name, None)
    if llm_config is None:
        raise ValueError(f"No config array was found for {vector_name} in {filename}")
    
    log.info(f'llm_config: {llm_config} for {vector_name} - fetching "{key}"')

    key_value = llm_config.get(key, None)
    
    return key_value


import os
from urllib.parse import quote
from datetime import datetime, timedelta
from google.cloud import storage 
import google.auth
from google.auth.transport.requests import Request
from google.auth.exceptions import RefreshError

from ..logging import log

# Perform a refresh request to get the access token of the current credentials (Else, it's None)
# Initialize global variables for client reuse
gcs_credentials, project_id = google.auth.default()
gcs_client = storage.Client(credentials=gcs_credentials)
gcs_bucket_cache = {}

# Prepare global variables for client reuse
def get_bucket(bucket_name):
    if bucket_name not in gcs_bucket_cache:
        gcs_bucket_cache[bucket_name] = gcs_client.get_bucket(bucket_name)
    return gcs_bucket_cache[bucket_name]

def sign_gcs_url(bucket_name:str, object_name:str, expiry_secs=86400):
    try:
        bucket = get_bucket(bucket_name)
        blob = bucket.blob(object_name)
        expires = datetime.now() + timedelta(seconds=expiry_secs)
        service_account_email = getattr(gcs_credentials, "service_account_email", os.getenv('GCS_MAIL_USER'))

        if not service_account_email:
            log.error("For local testing must set a GCS_MAIL_USER to sign GCS URLs")
            return None

        return blob.generate_signed_url(
            version="v4",
            expiration=expires,
            service_account_email=service_account_email,
            access_token=gcs_credentials.token
        )
    except RefreshError:
        log.info("Refreshing gcs_credentials due to token expiration.")
        gcs_credentials.refresh(Request())
        return sign_gcs_url(bucket_name, object_name, expiry_secs)
    except Exception as e:
        log.error(f"Failed to generate signed URL: {e}")
        return None


def construct_download_link(source_uri: str) -> str:
    """Creates a viewable Cloud Storage web browser link from a gs:// URI.""" 
    if not source_uri.startswith("gs://"):
        return source_uri, source_uri, False  # Return the URI as is if it doesn't start with gs://

    bucket_name, object_name = parse_gs_uri(source_uri)

    signed_url = sign_gcs_url(bucket_name, object_name)
    if signed_url:
        the_name = os.path.basename(object_name)
        log.info(f"Creating signed URL for {the_name} - {signed_url}")
        return signed_url, the_name, True
    
    log.error(f"Failed to generate signed URL for {source_uri}")
    return construct_download_link_simple(bucket_name, object_name)


def construct_download_link_simple(bucket_name:str, object_name:str) -> str:
    """Creates a viewable Cloud Storage web browser link from a gs:// URI.

    Args:
        source_uri: The gs:// URI of the object in Cloud Storage.

    Returns:
        A URL that directly access the object in the Cloud Storage web browser.
    """

    public_url = f"https://storage.cloud.google.com/{bucket_name}/{quote(object_name)}"
    filename = os.path.basename(object_name)
    signed = False
    return public_url, filename, signed

def parse_gs_uri(gs_uri: str) -> tuple[str, str]:
    """Parses a gs:// URI into the bucket name and object name.

    Args:
        gs_uri: The gs:// URI to parse.

    Returns:
        A tuple containing the bucket name and object name.
    """
    parts = gs_uri.split("/")
    if len(parts) < 3:
        raise ValueError(f"Invalid gs:// URI: {gs_uri}")
    return parts[2], "/".join(parts[3:])